@@ -29,6 +29,52 @@
 #include "pango-language.h"
 #include "pango-impl-utils.h"
 
+
+/* We embed a private struct right *before* a where a PangoLanguage *
+ * points to.
+ */
+
+typedef struct {
+  gconstpointer lang_info;
+  gconstpointer script_for_lang;
+
+  int magic; /* Used for verification */
+} PangoLanguagePrivate;
+
+#define PANGO_LANGUAGE_PRIVATE_MAGIC 0x0BE4DAD0
+
+static void
+pango_language_private_init (PangoLanguagePrivate *priv)
+{
+  priv->magic = PANGO_LANGUAGE_PRIVATE_MAGIC;
+
+  priv->lang_info = (gconstpointer) -1;
+  priv->script_for_lang = (gconstpointer) -1;
+}
+
+static PangoLanguagePrivate * pango_language_get_private (PangoLanguage *language) G_GNUC_CONST;
+
+static PangoLanguagePrivate *
+pango_language_get_private (PangoLanguage *language)
+{
+  PangoLanguagePrivate *priv;
+
+  if (!language)
+    return NULL;
+
+  priv = (PangoLanguagePrivate *) ((char *)language - sizeof (PangoLanguagePrivate));
+
+  if (G_UNLIKELY (priv->magic != PANGO_LANGUAGE_PRIVATE_MAGIC))
+    {
+      g_critical ("Invalid PangoLanguage.  Did you pass in a straight string instead of calling pango_language_from_string()?");
+      return NULL;
+    }
+
+  return priv;
+}
+
+
+
 #define LANGUAGE_SEPARATORS ";:, \t"
 
 static const char canon_map[256] = {
@@ -229,6 +275,7 @@ PangoLanguage *
 pango_language_from_string (const char *language)
 {
   static GHashTable *hash = NULL;
+  PangoLanguagePrivate *priv;
   char *result;
   int len;
   char *p;
@@ -246,7 +293,13 @@ pango_language_from_string (const char *language)
     }
 
   len = strlen (language);
-  result = g_malloc (len + 1);
+  result = g_malloc0 (sizeof (PangoLanguagePrivate) + len + 1);
+  g_assert (result);
+
+  priv = (PangoLanguagePrivate *) result;
+  result += sizeof (*priv);
+
+  pango_language_private_init (priv);
 
   p = result;
   while ((*(p++) = canon_map[*(guchar *)language++]))
@@ -318,11 +371,6 @@ pango_language_matches (PangoLanguage *language,
   return FALSE;
 }
 
-typedef struct {
-  const char lang[4];
-  const char *str;
-} LangInfo;
-
 static int
 lang_compare_first_component (gconstpointer pa,
 			      gconstpointer pb)
@@ -340,45 +388,121 @@ lang_compare_first_component (gconstpointer pa,
   return strncmp (a, b, MAX (da, db));
 }
 
-static int
-lang_info_compare (gconstpointer key,
-		   gconstpointer val)
+/* Finds the best record for @language in an array of records.
+ * Each record should start with the string representation of the language
+ * code for the record (embedded, not a pointer), and the records must be
+ * sorted on language code.
+ */
+static gconstpointer
+find_best_lang_match (PangoLanguage *language,
+		      gconstpointer  records,
+		      guint          num_records,
+		      guint          record_size)
 {
-  const LangInfo *lang_info = val;
+  const char *lang_str;
+  const char *record, *start, *end;
 
-  return lang_compare_first_component (key, lang_info->lang);
+  if (language == NULL)
+    return NULL;
+
+  lang_str = pango_language_to_string (language);
+
+  record = bsearch (lang_str,
+		    records, num_records, record_size,
+		    lang_compare_first_component);
+  if (!record)
+    return NULL;
+
+  start = (const char *) records;
+  end   = start + num_records * record_size;
+
+  /* find the best match among all those that have the same first-component */
+
+  /* go to the final one matching in the first component */
+  while (record < end - record_size &&
+	 lang_compare_first_component (lang_str, record + record_size) == 0)
+    record += record_size;
+
+  /* go back, find which one matches completely */
+  while (start <= record &&
+	 lang_compare_first_component (lang_str, record) == 0)
+    {
+      if (pango_language_matches (language, record))
+        return record;
+
+      record -= record_size;
+    }
+
+  return NULL;
 }
 
-/* The following array is supposed to contain enough text to tickle all necessary fonts for each
- * of the languages in the following. Yes, it's pretty lame. Not all of the languages
- * in the following have sufficient text to exercise all the accents for the language, and
- * there are obviously many more languages to include as well.
- */
+static gconstpointer
+find_best_lang_match_cached (PangoLanguage *language,
+			     gconstpointer *cache,
+			     gconstpointer  records,
+			     guint          num_records,
+			     guint          record_size)
+{
+  gconstpointer result;
+
+  if (G_LIKELY (cache && *cache != (gconstpointer) -1))
+    return *cache;
+
+  result = find_best_lang_match (language,
+				 records,
+				 num_records,
+				 record_size);
+
+  if (cache)
+    *cache = result;
+
+  return result;
+}
+
+#define FIND_BEST_LANG_MATCH(language, records) \
+	find_best_lang_match ((language), \
+			      records, \
+			      G_N_ELEMENTS (records), \
+			      sizeof (*records));
+
+#define FIND_BEST_LANG_MATCH_CACHED(language, cache_key, records) \
+	find_best_lang_match_cached ((language), \
+				     pango_language_get_private (language) ? \
+				       &(pango_language_get_private (language)->cache_key) : NULL, \
+				     records, \
+				     G_N_ELEMENTS (records), \
+				     sizeof (*records));
+
+typedef struct {
+  char lang[6];
+  guint16 offset;
+} LangInfo;
+
+/* Pure black magic, based on appendix of dsohowto.pdf */
+#define POOLSTRFIELD(line) POOLSTRFIELD1(line)
+#define POOLSTRFIELD1(line) str##line
+struct _LangPoolStruct {
+  char str0[1];
+#define LANGUAGE(id, source, sample) char POOLSTRFIELD(__LINE__)[sizeof(sample)];
+#include "pango-language-sample-table.h"
+#undef LANGUAGE
+};
+
+static const union _LangPool {
+  struct _LangPoolStruct lang_pool_struct;
+  const char str[1];
+} lang_pool = { {
+    "",
+#define LANGUAGE(id, source, sample) sample,
+#include "pango-language-sample-table.h"
+#undef LANGUAGE
+} };
 static const LangInfo lang_texts[] = {
-  { "ar", "Arabic  \330\247\331\204\330\263\331\204\330\247\331\205 \330\271\331\204\331\212\331\203\331\205" },
-  { "cs", "Czech (\304\215esky)  Dobr\303\275 den" },
-  { "da", "Danish (Dansk)  Hej, Goddag" },
-  { "el", "Greek (\316\225\316\273\316\273\316\267\316\275\316\271\316\272\316\254) \316\223\316\265\316\271\316\254 \317\203\316\261\317\202" },
-  { "en", "English Hello" },
-  { "eo", "Esperanto Saluton" },
-  { "es", "Spanish (Espa\303\261ol) \302\241Hola!" },
-  { "et", "Estonian  Tere, Tervist" },
-  { "fi", "Finnish (Suomi)  Hei, Hyv\303\244\303\244 p\303\244iv\303\244\303\244" },
-  { "fr", "French (Fran\303\247ais)" },
-  { "de", "German Gr\303\274\303\237 Gott" },
-  { "he", "Hebrew   \327\251\327\234\327\225\327\235" },
-  { "it", "Italiano  Ciao, Buon giorno" },
-  { "ja", "Japanese (\346\227\245\346\234\254\350\252\236) \343\201\223\343\202\223\343\201\253\343\201\241\343\201\257, \357\275\272\357\276\235\357\276\206\357\276\201\357\276\212" },
-  { "ko", "Korean (\355\225\234\352\270\200)   \354\225\210\353\205\225\355\225\230\354\204\270\354\232\224, \354\225\210\353\205\225\355\225\230\354\213\255\353\213\210\352\271\214" },
-  { "mt", "Maltese   \304\212aw, Sa\304\247\304\247a" },
-  { "nl", "Nederlands, Vlaams Hallo, Dag" },
-  { "no", "Norwegian (Norsk) Hei, God dag" },
-  { "pl", "Polish   Dzie\305\204 dobry, Hej" },
-  { "ru", "Russian (\320\240\321\203\321\201\321\201\320\272\320\270\320\271)" },
-  { "sk", "Slovak   Dobr\303\275 de\305\210" },
-  { "sv", "Swedish (Svenska) Hej p\303\245 dej, Goddag" },
-  { "tr", "Turkish (T\303\274rk\303\247e) Merhaba" },
-  { "zh", "Chinese (\344\270\255\346\226\207,\346\231\256\351\200\232\350\257\235,\346\261\211\350\257\255)" }
+#define LANGUAGE(id, source, sample) {G_STRINGIFY(id),	G_STRUCT_OFFSET(struct _LangPoolStruct, POOLSTRFIELD(__LINE__))},
+#include "pango-language-sample-table.h"
+#undef LANGUAGE
+  /* One extra entry with no final comma, to make it C89-happy */
+ {"~~",	0}
 };
 
 /**
@@ -386,51 +510,119 @@ static const LangInfo lang_texts[] = {
  * @language: a #PangoLanguage, or %NULL
  *
  * Get a string that is representative of the characters needed to
- * render a particular language. This function is a bad hack for
- * internal use by renderers and Pango.
+ * render a particular language.
+ *
+ * The sample text may be a pangram, but is not necessarily.  It is chosen to
+ * be demonstrative of normal text in the language, as well as exposing font
+ * feature requirements unique to the language.  It is suitable for use
+ * as sample text in a font selection dialog.
+ *
+ * If @language is %NULL, the default language as found by
+ * pango_language_get_default() is used.
+ *
+ * If Pango does not have a sample string for @language, the classic
+ * "The quick brown fox..." is returned.  This can be detected by
+ * comparing the returned pointer value to that returned for (non-existent)
+ * language code "xx".  That is, compare to:
+ * <informalexample><programlisting>
+ * pango_language_get_sample_string (pango_language_from_string ("xx"))
+ * </programlisting></informalexample>
  *
  * Return value: the sample string. This value is owned by Pango
- *   and must not be freed.
+ *   and should not be freed.
  **/
 G_CONST_RETURN char *
 pango_language_get_sample_string (PangoLanguage *language)
 {
-  const char *result;
+  const LangInfo *lang_info;
 
-  if (language)
-    {
-      const char *lang_str = pango_language_to_string (language);
+  if (!language)
+    language = pango_language_get_default ();
 
-      LangInfo *lang_info = bsearch (lang_str, lang_texts,
-				     G_N_ELEMENTS (lang_texts), sizeof (LangInfo),
-				     lang_info_compare);
+  lang_info = FIND_BEST_LANG_MATCH_CACHED (language,
+					   lang_info,
+					   lang_texts);
 
-      if (lang_info)
-	result = lang_info->str;
-      else
-	result = "French (Fran\303\247ais)";     /* Assume iso-8859-1 */
-    }
-  else
-    {
-      /* Complete junk
-       */
+  if (lang_info)
+    return lang_pool.str + lang_info->offset;
 
-      result = "\330\247\331\204\330\263\331\204\330\247\331\205 \330\271\331\204\331\212\331\203\331\205 \304\215esky \316\225\316\273\316\273\316\267\316\275\316\271\316\272\316\254 Fran\303\247ais \346\227\245\346\234\254\350\252\236 \355\225\234\352\270\200 \320\240\321\203\321\201\321\201\320\272\320\270\320\271 \344\270\255\346\226\207,\346\231\256\351\200\232\350\257\235,\346\261\211\350\257\255 T\303\274rk\303\247e";
-    }
-
-  return result;
+  return "The quick brown fox jumps over the lazy dog.";
 }
 
+
+
+
+/*
+ * From language to script
+ */
+
+
 #include "pango-script-lang-table.h"
 
-static int
-script_for_lang_compare (gconstpointer key,
-			 gconstpointer member)
+/**
+ * pango_language_get_scripts:
+ * @language: a #PangoLanguage, or %NULL
+ * @num_scripts: location to return number of scripts, or %NULL
+ *
+ * Determines the scripts used to to write @language.
+ * If nothing is known about the language tag @language,
+ * or if @language is %NULL, then %NULL is returned.
+ * The list of scripts returned starts with the script that the
+ * language uses most and continues to the one it uses least.
+ *
+ * The value @num_script points at will be set to the number
+ * of scripts in the returned array (or zero if %NULL is returned).
+ *
+ * Most languages use only one script for writing, but there are
+ * some that use two (Latin and Cyrillic for example), and a few
+ * use three (Japanese for example).  Applications should not make
+ * any assumptions on the maximum number of scripts returned
+ * though, except that it is positive if the return value is not
+ * %NULL, and it is a small number.
+ *
+ * The pango_language_includes_script() function uses this function
+ * internally.
+ *
+ * Return value: An array of #PangoScript values, with the
+ * number of entries in the array stored in @num_scripts, or
+ * %NULL if Pango does not have any information about this
+ * particular language tag (also the case if @language is %NULL).
+ * The returned array is owned by Pango and should not be modified
+ * or freed.
+ 
+ * Since: 1.22
+ **/
+G_CONST_RETURN PangoScript *
+pango_language_get_scripts (PangoLanguage *language,
+			    int           *num_scripts)
 {
-  PangoLanguage *lang = (PangoLanguage *)key;
-  const PangoScriptForLang *script_for_lang = member;
+  const PangoScriptForLang *script_for_lang;
+  unsigned int j;
 
-  return lang_compare_first_component (lang, script_for_lang->lang);
+  script_for_lang = FIND_BEST_LANG_MATCH_CACHED (language,
+						 script_for_lang,
+						 pango_script_for_lang);
+
+  if (!script_for_lang)
+    {
+      if (num_scripts)
+	*num_scripts = 0;
+
+      return NULL;
+    }
+
+  if (num_scripts)
+    {
+      for (j = 0; j < G_N_ELEMENTS (script_for_lang->scripts); j++)
+	if (script_for_lang->scripts[j] == 0)
+	  break;
+
+      g_assert (j > 0);
+
+      *num_scripts = j;
+    }
+
+  return script_for_lang->scripts;
 }
 
 /**
@@ -449,6 +641,8 @@ script_for_lang_compare (gconstpointer key,
  * a particular section of text. It probably is not useful for
  * applications in most circumstances.
  *
+ * This function uses pango_language_get_scripts() internally.
+ *
  * Return value: %TRUE if @script is one of the scripts used
  * to write @language or if nothing is known about @language
  * (including the case that @language is %NULL),
@@ -460,64 +654,37 @@ gboolean
 pango_language_includes_script (PangoLanguage *language,
 				PangoScript    script)
 {
-  PangoScriptForLang *script_for_lang;
-  unsigned int j;
-  const char *lang_str;
+  const PangoScript *scripts;
+  int num_scripts, j;
 
+/* copied from the one in pango-script.c */
 #define REAL_SCRIPT(script) \
-  ((script) > PANGO_SCRIPT_INHERITED)
+  ((script) > PANGO_SCRIPT_INHERITED && (script) != PANGO_SCRIPT_UNKNOWN)
 
-  if (language == NULL || !REAL_SCRIPT (script))
+  if (!REAL_SCRIPT (script))
     return TRUE;
 
-  lang_str = pango_language_to_string (language);
+#undef REAL_SCRIPT
 
-  /* This bsearch could be optimized to occur only once if
-   * we store the pointer to the PangoScriptForLang in the
-   * same block as the string value for the PangoLanguage.
-   */
-  script_for_lang = bsearch (lang_str,
-			     pango_script_for_lang,
-			     G_N_ELEMENTS (pango_script_for_lang),
-			     sizeof (PangoScriptForLang),
-			     script_for_lang_compare);
-  if (!script_for_lang)
+  scripts = pango_language_get_scripts (language, &num_scripts);
+  if (!scripts)
     return TRUE;
-  else
-    {
-      gboolean found = FALSE;
 
-      /* find the best matching language */
-     
-      /* go to the final one matching in the first component */
-      while (script_for_lang + 1 < pango_script_for_lang + G_N_ELEMENTS (pango_script_for_lang) &&
-	     script_for_lang_compare (lang_str, script_for_lang + 1) == 0)
-        script_for_lang++;
-
-      /* go back, find which one matches completely */
-      while (script_for_lang >= pango_script_for_lang &&
-	     script_for_lang_compare (lang_str, script_for_lang) == 0)
-        {
-	  if (pango_language_matches (language, script_for_lang->lang))
-	    {
-	      found = TRUE;
-	      break;
-	    }
-
-          script_for_lang--;
-	}
-
-      if (!found)
-        return TRUE;
-    }
-
-  for (j = 0; j < G_N_ELEMENTS (script_for_lang->scripts); j++)
-    if (script_for_lang->scripts[j] == script)
+  for (j = 0; j < num_scripts; j++)
+    if (scripts[j] == script)
       return TRUE;
 
   return FALSE;
 }
 
+
+
+
+/*
+ * From script to language
+ */
+
+
 static PangoLanguage **
 parse_default_languages (void)
 {
@@ -694,11 +861,10 @@ pango_script_get_sample_language (PangoScript script)
     "iu",  /* PANGO_SCRIPT_CANADIAN_ABORIGINAL */
     "",    /* PANGO_SCRIPT_YI */
     "tl",  /* PANGO_SCRIPT_TAGALOG */
-    /* There are no ISO-636 language codes for the following
-     * Phillipino languages/scripts */
-    "",    /* PANGO_SCRIPT_HANUNOO */
-    "",    /* PANGO_SCRIPT_BUHID */
-    "",    /* PANGO_SCRIPT_TAGBANWA */
+    /* Phillipino languages/scripts */
+    "hnn", /* PANGO_SCRIPT_HANUNOO */
+    "bku", /* PANGO_SCRIPT_BUHID */
+    "tbw", /* PANGO_SCRIPT_TAGBANWA */
 
     "",    /* PANGO_SCRIPT_BRAILLE */
     "",    /* PANGO_SCRIPT_CYPRIOT */
@@ -718,8 +884,7 @@ pango_script_get_sample_language (PangoScript script)
     "",    /* PANGO_SCRIPT_GLAGOLITIC */
     /* Used for for Berber (ber), but Arabic script is more common */
     "",    /* PANGO_SCRIPT_TIFINAGH */
-    /* Syloti Nagri is used for Sylheti, no ISO 639 code */
-    "",    /* PANGO_SCRIPT_SYLOTI_NAGRI */
+    "syl", /* PANGO_SCRIPT_SYLOTI_NAGRI */
     "peo", /* PANGO_SCRIPT_OLD_PERSIAN */
     "",    /* PANGO_SCRIPT_KHAROSHTHI */
 
