@@ -19,7 +19,7 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include <config.h>
+#include "config.h"
 
 #include "pango-ot-private.h"
 #include "pango-impl-utils.h"
@@ -29,7 +29,7 @@ typedef struct _PangoOTRule PangoOTRule;
 struct _PangoOTRule
 {
   gulong     property_bit;
-  FT_UShort  feature_index;
+  HB_UShort  feature_index;
   guint      table_type : 1;
 };
 
@@ -93,7 +93,7 @@ pango_ot_ruleset_finalize (GObject *object)
 
   g_array_free (ruleset->rules, TRUE);
   if (ruleset->info)
-    g_object_remove_weak_pointer (ruleset->info, &ruleset->info);
+    g_object_remove_weak_pointer (G_OBJECT (ruleset->info), (gpointer *) &ruleset->info);
 
   parent_class->finalize (object);
 }
@@ -175,7 +175,7 @@ pango_ot_ruleset_new (PangoOTInfo *info)
   ruleset = g_object_new (PANGO_TYPE_OT_RULESET, NULL);
 
   ruleset->info = info;
-  g_object_add_weak_pointer (ruleset->info, &ruleset->info);
+  g_object_add_weak_pointer (G_OBJECT (ruleset->info), (gpointer *) &ruleset->info);
 
   return ruleset;
 }
@@ -414,7 +414,7 @@ pango_ot_ruleset_maybe_add_feature (PangoOTRuleset          *ruleset,
  *
  * This is a convenience function that 
  * for each feature in the feature map array @features
- * converts the feature name to a #PangoOTTag feature tag using FT_MAKE_TAG()
+ * converts the feature name to a #PangoOTTag feature tag using PANGO_OT_TAG_MAKE()
  * and calls pango_ot_ruleset_maybe_add_feature() on it.
  *
  * Return value: The number of features in @features that were found
@@ -435,10 +435,10 @@ pango_ot_ruleset_maybe_add_features (PangoOTRuleset          *ruleset,
 
   for (i = 0; i < n_features; i++)
     {
-      PangoOTTag feature_tag = FT_MAKE_TAG (features[i].feature_name[0],
-					    features[i].feature_name[1],
-					    features[i].feature_name[2],
-					    features[i].feature_name[3]);
+      PangoOTTag feature_tag = PANGO_OT_TAG_MAKE (features[i].feature_name[0],
+						  features[i].feature_name[1],
+						  features[i].feature_name[2],
+						  features[i].feature_name[3]);
 
       n_found_features += pango_ot_ruleset_maybe_add_feature (ruleset,
 							      table_type,
@@ -564,7 +564,7 @@ pango_ot_ruleset_position (const PangoOTRuleset *ruleset,
 
   if (HB_GPOS_Apply_String (ruleset->info->face, gpos, 0, buffer->buffer,
 			    FALSE /* enable device-dependant values */,
-			    buffer->rtl) == FT_Err_Ok)
+			    buffer->rtl) == HB_Err_Ok)
     buffer->applied_gpos = TRUE;
 }
 
