@@ -20,10 +20,11 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include <config.h>
+#include "config.h"
 
 #import <Cocoa/Cocoa.h>
 
+#include "pango-impl-utils.h"
 #include "pangoatsui-private.h"
 #include "pangocairo.h"
 #include "pangocairo-private.h"
@@ -148,7 +149,7 @@ pango_cairo_atsui_font_create_metrics_for_context (PangoCairoFont *font,
   pango_layout_set_text (layout, sample_str, -1);
   pango_layout_get_extents (layout, NULL, &extents);
 
-  metrics->approximate_char_width = extents.width / g_utf8_strlen (sample_str, -1);
+  metrics->approximate_char_width = extents.width / pango_utf8_strwidth (sample_str);
 
   pango_layout_set_text (layout, "0123456789", -1);
   metrics->approximate_digit_width = max_glyph_width (layout);
@@ -174,6 +175,10 @@ pango_cairo_atsui_font_describe_absolute (PangoFont *font)
 static void
 pango_cairo_atsui_font_finalize (GObject *object)
 {
+  PangoCairoATSUIFont *cafont = (PangoCairoATSUIFont *) object;
+
+  _pango_cairo_font_private_finalize (&cafont->cf_priv);
+
   G_OBJECT_CLASS (pango_cairo_atsui_font_parent_class)->finalize (object);
 }
 
@@ -191,7 +196,7 @@ pango_cairo_atsui_font_class_init (PangoCairoATSUIFontClass *class)
 }
 
 static void
-pango_cairo_atsui_font_init (PangoCairoATSUIFont *cafont)
+pango_cairo_atsui_font_init (PangoCairoATSUIFont *cafont G_GNUC_UNUSED)
 {
 }
 
