@@ -20,7 +20,7 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include <config.h>
+#include "config.h"
 #include <math.h>
 
 #include "pangoft2-private.h"
@@ -96,24 +96,124 @@ static void
 pango_ft2_free_rendered_glyph (PangoFT2RenderedGlyph *rendered)
 {
   g_free (rendered->bitmap.buffer);
-  g_free (rendered);
+  g_slice_free (PangoFT2RenderedGlyph, rendered);
+}
+
+static PangoFT2RenderedGlyph *
+pango_ft2_font_render_box_glyph (int      width,
+				 int      height,
+				 int      top,
+				 gboolean invalid)
+{
+  PangoFT2RenderedGlyph *box;
+  int i, j, offset1, offset2, line_width;
+
+  line_width = MAX ((height + 43) / 44, 1);
+  if (width < 1 || height < 1)
+    line_width = 0;
+
+  box = g_slice_new (PangoFT2RenderedGlyph);
+
+  box->bitmap_left = 0;
+  box->bitmap_top = top;
+
+  box->bitmap.pixel_mode = ft_pixel_mode_grays;
+
+  box->bitmap.width = width;
+  box->bitmap.rows = height;
+  box->bitmap.pitch = height;
+
+  box->bitmap.buffer = g_malloc0 (box->bitmap.rows * box->bitmap.pitch);
+
+  /* draw the box */
+  for (j = 0; j < line_width; j++)
+    {
+      offset1 = box->bitmap.pitch * (MIN (1 + j, height - 1));
+      offset2 = box->bitmap.pitch * (MAX (box->bitmap.rows - 2 - j, 0));
+      for (i = 1;
+	   i < box->bitmap.width - 1;
+	   i++)
+	{
+	  box->bitmap.buffer[offset1 + i] = 0xff;
+	  box->bitmap.buffer[offset2 + i] = 0xff;
+	}
+    }
+  for (j = 0; j < line_width; j++)
+    {
+      offset1 = MIN (1 + j, width - 1);
+      offset2 = MAX (box->bitmap.width - 2 - j, 0);
+      for (i = box->bitmap.pitch;
+	   i < (box->bitmap.rows - 1) * box->bitmap.pitch;
+	   i += box->bitmap.pitch)
+	{
+	  box->bitmap.buffer[offset1 + i] = 0xff;
+	  box->bitmap.buffer[offset2 + i] = 0xff;
+	}
+    }
+
+  if (invalid)
+    {
+      /* XXX This may scrabble memory.  Didn't check close enough */
+      int inc = PANGO_SCALE * MAX (width - line_width, 0) / (height + 1);
+      offset1 = PANGO_SCALE;
+      offset2 = PANGO_SCALE * MAX (width - line_width - 1, 0) ;
+      for (i = box->bitmap.pitch;
+	   i < (box->bitmap.rows - 1) * box->bitmap.pitch;
+	   i += box->bitmap.pitch)
+        {
+	  for (j = 0; j < line_width; j++)
+	    {
+	      box->bitmap.buffer[PANGO_PIXELS (offset1) + i + j] = 0xff;
+	      box->bitmap.buffer[PANGO_PIXELS (offset2) + i + j] = 0xff;
+	    }
+	  offset1 += inc;
+	  offset2 -= inc;
+	}
+
+    }
+
+  return box;
 }
 
 static PangoFT2RenderedGlyph *
 pango_ft2_font_render_glyph (PangoFont *font,
-			     int glyph_index)
+			     PangoGlyph glyph_index)
 {
-  PangoFT2RenderedGlyph *rendered;
   FT_Face face;
+  gboolean invalid_input;
 
-  rendered = g_new (PangoFT2RenderedGlyph, 1);
+  invalid_input = glyph_index == PANGO_GLYPH_INVALID_INPUT || (glyph_index & ~PANGO_GLYPH_UNKNOWN_FLAG) > 0x10FFFF;
+
+  if (glyph_index & PANGO_GLYPH_UNKNOWN_FLAG)
+    {
+      PangoFT2RenderedGlyph *box;
+      PangoFontMetrics *metrics;
+
+      if (!font)
+	goto generic_box;
+
+      metrics = pango_font_get_metrics (font, NULL);
+      if (!metrics)
+	goto generic_box;
+
+      box = pango_ft2_font_render_box_glyph (PANGO_PIXELS (metrics->approximate_char_width),
+					     PANGO_PIXELS (metrics->ascent + metrics->descent),
+					     PANGO_PIXELS (metrics->ascent),
+					     invalid_input);
+      pango_font_metrics_unref (metrics);
+
+      return box;
+    }
 
   face = pango_ft2_font_get_face (font);
-  
+
   if (face)
     {
+      PangoFT2RenderedGlyph *rendered;
       PangoFT2Font *ft2font = (PangoFT2Font *) font;
 
+      rendered = g_slice_new (PangoFT2RenderedGlyph);
+
       /* Draw glyph */
       FT_Load_Glyph (face, glyph_index, ft2font->load_flags);
       FT_Render_Glyph (face->glyph,
@@ -125,11 +225,17 @@ pango_ft2_font_render_glyph (PangoFont *font,
 					  face->glyph->bitmap.rows * face->glyph->bitmap.pitch);
       rendered->bitmap_left = face->glyph->bitmap_left;
       rendered->bitmap_top = face->glyph->bitmap_top;
+
+      return rendered;
     }
   else
-    g_error ("Couldn't get face for PangoFT2Face");
-
-  return rendered;
+    {
+generic_box:
+      return  pango_ft2_font_render_box_glyph (PANGO_UNKNOWN_GLYPH_WIDTH,
+					       PANGO_UNKNOWN_GLYPH_HEIGHT,
+					       PANGO_UNKNOWN_GLYPH_HEIGHT,
+					       invalid_input);
+    }
 }
 
 static void
@@ -150,6 +256,21 @@ pango_ft2_renderer_draw_glyph (PangoRenderer *renderer,
   int iyoff = floor (y + 0.5);
   int ix, iy;
 
+  if (glyph & PANGO_GLYPH_UNKNOWN_FLAG)
+    {
+      /* Since we don't draw hexbox for FT2 renderer,
+       * unifiy the rendered bitmap in the cache by converting
+       * all missing glyphs to either INVALID_INPUT or UNKNOWN_FLAG.
+       */
+
+      gunichar wc = glyph & (~PANGO_GLYPH_UNKNOWN_FLAG);
+
+      if (G_UNLIKELY (glyph == PANGO_GLYPH_INVALID_INPUT || wc > 0x10FFFF))
+	glyph = PANGO_GLYPH_INVALID_INPUT;
+      else
+	glyph = PANGO_GLYPH_UNKNOWN_FLAG;
+    }
+
   rendered_glyph = _pango_ft2_font_get_cache_glyph_data (font, glyph);
   add_glyph_to_cache = FALSE;
   if (rendered_glyph == NULL)
@@ -215,7 +336,7 @@ pango_ft2_renderer_draw_glyph (PangoRenderer *renderer,
 	    {
 	      if ((*s) & (1 << (7 - (ix % 8))))
 		*d |= 0xff;
-		      
+
 	      if ((ix % 8) == 7)
 		s++;
 	      d++;
@@ -225,7 +346,7 @@ pango_ft2_renderer_draw_glyph (PangoRenderer *renderer,
 	  src  += rendered_glyph->bitmap.pitch;
 	}
       break;
-	      
+
     default:
       g_warning ("pango_ft2_render: "
 		 "Unrecognized glyph bitmap pixel mode %d\n",
@@ -443,7 +564,7 @@ pango_ft2_renderer_draw_trapezoid (PangoRenderer   *renderer,
 
 /**
  * pango_ft2_render_layout_subpixel:
- * @bitmap:    a FT_Bitmap to render the layout onto
+ * @bitmap:    a <type>FT_Bitmap</type> to render the layout onto
  * @layout:    a #PangoLayout
  * @x:         the X position of the left of the layout (in Pango units)
  * @y:         the Y position of the top of the layout (in Pango units)
@@ -456,10 +577,10 @@ pango_ft2_renderer_draw_trapezoid (PangoRenderer   *renderer,
  *
  * Since: 1.6
  */
-void 
+void
 pango_ft2_render_layout_subpixel (FT_Bitmap   *bitmap,
 				  PangoLayout *layout,
-				  int          x, 
+				  int          x,
 				  int          y)
 {
   PangoContext *context;
@@ -474,23 +595,23 @@ pango_ft2_render_layout_subpixel (FT_Bitmap   *bitmap,
   renderer = _pango_ft2_font_map_get_renderer (PANGO_FT2_FONT_MAP (fontmap));
 
   pango_ft2_renderer_set_bitmap (PANGO_FT2_RENDERER (renderer), bitmap);
-  
+
   pango_renderer_draw_layout (renderer, layout, x, y);
 }
 
 /**
  * pango_ft2_render_layout:
- * @bitmap:    a FT_Bitmap to render the layout onto
+ * @bitmap:    a <type>FT_Bitmap</type> to render the layout onto
  * @layout:    a #PangoLayout
  * @x:         the X position of the left of the layout (in pixels)
  * @y:         the Y position of the top of the layout (in pixels)
  *
  * Render a #PangoLayout onto a FreeType2 bitmap
  */
-void 
+void
 pango_ft2_render_layout (FT_Bitmap   *bitmap,
 			 PangoLayout *layout,
-			 int          x, 
+			 int          x,
 			 int          y)
 {
   pango_ft2_render_layout_subpixel (bitmap, layout, x * PANGO_SCALE, y * PANGO_SCALE);
@@ -498,7 +619,7 @@ pango_ft2_render_layout (FT_Bitmap   *bitmap,
 
 /**
  * pango_ft2_render_layout_line_subpixel:
- * @bitmap:    a FT_Bitmap to render the line onto
+ * @bitmap:    a <type>FT_Bitmap</type> to render the line onto
  * @line:      a #PangoLayoutLine
  * @x:         the x position of start of string (in Pango units)
  * @y:         the y position of baseline (in Pango units)
@@ -511,10 +632,10 @@ pango_ft2_render_layout (FT_Bitmap   *bitmap,
  *
  * Since: 1.6
  */
-void 
+void
 pango_ft2_render_layout_line_subpixel (FT_Bitmap       *bitmap,
 				       PangoLayoutLine *line,
-				       int              x, 
+				       int              x,
 				       int              y)
 {
   PangoContext *context;
@@ -529,23 +650,23 @@ pango_ft2_render_layout_line_subpixel (FT_Bitmap       *bitmap,
   renderer = _pango_ft2_font_map_get_renderer (PANGO_FT2_FONT_MAP (fontmap));
 
   pango_ft2_renderer_set_bitmap (PANGO_FT2_RENDERER (renderer), bitmap);
-  
+
   pango_renderer_draw_layout_line (renderer, line, x, y);
 }
 
 /**
  * pango_ft2_render_layout_line:
- * @bitmap:    a FT_Bitmap to render the line onto
+ * @bitmap:    a <type>FT_Bitmap</type> to render the line onto
  * @line:      a #PangoLayoutLine
  * @x:         the x position of start of string (in pixels)
  * @y:         the y position of baseline (in pixels)
  *
  * Render a #PangoLayoutLine onto a FreeType2 bitmap
  */
-void 
+void
 pango_ft2_render_layout_line (FT_Bitmap       *bitmap,
 			      PangoLayoutLine *line,
-			      int              x, 
+			      int              x,
 			      int              y)
 {
   pango_ft2_render_layout_line_subpixel (bitmap, line, x * PANGO_SCALE, y * PANGO_SCALE);
@@ -555,7 +676,7 @@ pango_ft2_render_layout_line (FT_Bitmap       *bitmap,
  * pango_ft2_render_transformed:
  * @bitmap:  the FreeType2 bitmap onto which to draw the string
  * @font:    the font in which to draw the string
- * @matrix: a #PangoMatrix, or %NULL to use an identity transformation
+ * @matrix:  a #PangoMatrix, or %NULL to use an identity transformation
  * @glyphs:  the glyph string to draw
  * @x:       the x position of the start of the string (in Pango
  *           units in user space coordinates)
@@ -571,27 +692,27 @@ pango_ft2_render_layout_line (FT_Bitmap       *bitmap,
  *
  * Since: 1.6
  **/
-void 
+void
 pango_ft2_render_transformed (FT_Bitmap         *bitmap,
 			      const PangoMatrix *matrix,
 			      PangoFont         *font,
 			      PangoGlyphString  *glyphs,
-			      int                x, 
+			      int                x,
 			      int                y)
 {
   PangoFontMap *fontmap;
   PangoRenderer *renderer;
 
   g_return_if_fail (bitmap != NULL);
-  g_return_if_fail (PANGO_FT2_IS_FONT (font));
   g_return_if_fail (glyphs != NULL);
+  g_return_if_fail (PANGO_FT2_IS_FONT (font));
 
   fontmap = PANGO_FC_FONT (font)->fontmap;
   renderer = _pango_ft2_font_map_get_renderer (PANGO_FT2_FONT_MAP (fontmap));
 
   pango_ft2_renderer_set_bitmap (PANGO_FT2_RENDERER (renderer), bitmap);
   pango_renderer_set_matrix (renderer, matrix);
-  
+
   pango_renderer_draw_glyphs (renderer, font, glyphs, x, y);
 }
 
@@ -605,11 +726,11 @@ pango_ft2_render_transformed (FT_Bitmap         *bitmap,
  *
  * Renders a #PangoGlyphString onto a FreeType2 bitmap.
  **/
-void 
+void
 pango_ft2_render (FT_Bitmap        *bitmap,
 		  PangoFont        *font,
 		  PangoGlyphString *glyphs,
-		  int               x, 
+		  int               x,
 		  int               y)
 {
   pango_ft2_render_transformed (bitmap, NULL, font, glyphs, x * PANGO_SCALE, y * PANGO_SCALE);
